import asyncio
import itertools

import discord
from discord.ext import commands

from plugins.core.navigation.paginator import Pages


class HelpPaginator(Pages):
    def __init__(self, help_command, ctx, entries, *, per_page=4):
        super().__init__(ctx, entries=entries, per_page=per_page)
        self.reaction_emojis.append(
            ("\N{WHITE QUESTION MARK ORNAMENT}", self.show_bot_help)
        )
        self.total = len(entries)
        self.help_command = help_command
        self.prefix = help_command.clean_prefix
        self.is_bot = False

    def get_bot_page(self, page):
        cog, description, commands = self.entries[page - 1]
        self.title = f"{cog} Commands"
        self.description = description
        return commands

    def prepare_embed(self, entries, page, *, first=False):
        self.embed.clear_fields()
        self.embed.description = self.description
        self.embed.title = self.title

        if self.is_bot:
            value = self.bot.config["help"]["support_text"]
            self.embed.add_field(name="Support", value=value, inline=False)

        self.embed.set_footer(
            text=f'Use "{self.prefix}help command" for more info on a command.'
        )

        for entry in entries:
            signature = f"{entry.qualified_name} {entry.signature}"
            self.embed.add_field(
                name=signature, value=entry.short_doc or "No help given", inline=False
            )

        self.embed.add_field(
            name="​",
            value="Powered by [Erin]" "(https://github.com/OpenDebates/Erin)",
        )

        if self.maximum_pages:
            self.embed.set_author(
                name=f"Page {page}/{self.maximum_pages} ({self.total} commands)"
            )

    async def show_help(self):
        """shows this message"""

        self.embed.title = "Paginator help"
        self.embed.description = "Hello! Welcome to the help page."

        messages = [f"{emoji} {func.__doc__}" for emoji, func in self.reaction_emojis]
        self.embed.clear_fields()
        self.embed.add_field(
            name="What are these reactions for?",
            value="\n".join(messages),
            inline=False,
        )

        self.embed.set_footer(
            text=f"We were on page {self.current_page} before this message."
        )
        await self.message.edit(embed=self.embed)

        async def go_back_to_current_page():
            await asyncio.sleep(30.0)
            await self.show_current_page()

        self.bot.loop.create_task(go_back_to_current_page())

    async def show_bot_help(self):
        """shows how to use the bot"""

        self.embed.title = "Using the bot"
        self.embed.description = "Hello! Welcome to the help page."
        self.embed.clear_fields()

        entries = (
            ("<argument>", "This means the argument is __**required**__."),
            ("[argument]", "This means the argument is __**optional**__."),
            ("[A|B]", "This means the it can be __**either A or B**__."),
            (
                "[argument...]",
                "This means you can have multiple arguments.\n"
                "Now that you know the basics, it should be noted that...\n"
                "__**You do not type in the brackets!**__",
            ),
        )

        self.embed.add_field(
            name="How do I use this bot?",
            value="Reading the bot signature is pretty simple.",
        )

        for name, value in entries:
            self.embed.add_field(name=name, value=value, inline=False)

        self.embed.set_footer(
            text=f"We were on page {self.current_page} before this message."
        )
        await self.message.edit(embed=self.embed)

        async def go_back_to_current_page():
            await asyncio.sleep(30.0)
            await self.show_current_page()

        self.bot.loop.create_task(go_back_to_current_page())


class PaginatedHelpCommand(commands.HelpCommand):
    def __init__(self):
        super().__init__(
            command_attrs={
                "cooldown": commands.Cooldown(1, 3.0, commands.BucketType.member),
                "help": "Shows help about the bot, a command, or a category",
            }
        )

    async def on_help_command_error(self, ctx, error):
        if isinstance(error, commands.CommandInvokeError):
            await ctx.send(str(error.original))

    def get_command_signature(self, command):
        parent = command.full_parent_name
        if len(command.aliases) > 0:
            aliases = "|".join(command.aliases)
            fmt = f"[{command.name}|{aliases}]"
            if parent:
                fmt = f"{parent} {fmt}"
            alias = fmt
        else:
            alias = command.name if not parent else f"{parent} {command.name}"
        return f"{alias} {command.signature}"

    async def send_bot_help(self, mapping):
        def key(c):
            return c.cog_name or "\u200bNo Category"

        bot = self.context.bot
        entries = await self.filter_commands(bot.commands, sort=True, key=key)
        nested_pages = []
        per_page = 9
        total = 0

        for cog, commands in itertools.groupby(entries, key=key):
            commands = sorted(commands, key=lambda c: c.name)
            if len(commands) == 0:
                continue

            total += len(commands)
            actual_cog = bot.get_cog(cog)
            # get the description if it exists (and the cog is valid) or return Empty embed.
            description = (actual_cog and actual_cog.description) or discord.Embed.Empty
            nested_pages.extend(
                (cog, description, commands[i : i + per_page])
                for i in range(0, len(commands), per_page)
            )

        # a value of 1 forces the pagination session
        pages = HelpPaginator(self, self.context, nested_pages, per_page=1)

        # swap the get_page implementation to work with our nested pages.
        pages.get_page = pages.get_bot_page
        pages.is_bot = True
        pages.total = total
        await pages.paginate()

    async def send_cog_help(self, cog):
        entries = await self.filter_commands(cog.get_commands(), sort=True)
        pages = HelpPaginator(self, self.context, entries)
        pages.title = f"{cog.qualified_name} Commands"
        pages.description = cog.description

        await pages.paginate()

    def common_command_formatting(self, page_or_embed, command):
        page_or_embed.title = self.get_command_signature(command)
        if command.description:
            page_or_embed.description = f"{command.description}\n\n{command.help}"
        else:
            page_or_embed.description = command.help or "No help found..."

    async def send_command_help(self, command):
        # No pagination necessary for a single command.
        embed = discord.Embed(colour=self.context.bot.config["help"]["color"])
        self.common_command_formatting(embed, command)
        await self.context.send(embed=embed)

    async def send_group_help(self, group):
        subcommands = group.commands
        if len(subcommands) == 0:
            return await self.send_command_help(group)

        entries = await self.filter_commands(subcommands, sort=True)
        pages = HelpPaginator(self, self.context, entries)
        self.common_command_formatting(pages, group)
        await pages.paginate()


class HelpMenu(commands.Cog, name="Help"):
    def __init__(self, bot):
        self.bot = bot
        self.logger = self.bot.logger
        self._original_help_command = bot.help_command
        bot.help_command = PaginatedHelpCommand()
        bot.help_command.cog = self


def setup(bot):
    bot.add_cog(HelpMenu(bot))



import asyncio

import discord


class CannotPaginate(Exception):
    pass


class Pages:
    """Implements a paginator that queries the user for the
    pagination interface.
    Pages are 1-index based, not 0-index based.
    If the user does not reply within 2 minutes then the pagination
    interface exits automatically.

    Parameters
    ------------
    ctx: discord.ext.commands.Context
        The context of the command.
    entries: List[str]
        A list of entries to paginate.
    per_page: int
        How many entries show up per page.
    show_entry_count: bool
        Whether to show an entry count in the footer.

    Attributes
    -----------
    embed: discord.Embed
        The embed object that is being used to send pagination info.
        Feel free to modify this externally. Only the description,
        footer fields, and colour are internally modified.
    permissions: discord.Permissions
        Our permissions for the channel.
    """

    def __init__(self, ctx, *, entries, per_page=12, show_entry_count=True):
        self.bot = ctx.bot
        self.entries = entries
        self.message = ctx.message
        self.channel = ctx.channel
        self.author = ctx.author
        self.per_page = per_page
        pages, left_over = divmod(len(self.entries), self.per_page)
        if left_over:
            pages += 1
        self.maximum_pages = pages
        self.embed = discord.Embed(colour=self.bot.config["help"]["color"])
        self.paginating = len(entries) > per_page
        self.show_entry_count = show_entry_count
        self.reaction_emojis = [
            (
                "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
                self.first_page,
            ),
            ("\N{BLACK LEFT-POINTING TRIANGLE}", self.previous_page),
            ("\N{BLACK RIGHT-POINTING TRIANGLE}", self.next_page),
            (
                "\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
                self.last_page,
            ),
            ("\N{INPUT SYMBOL FOR NUMBERS}", self.numbered_page),
            ("\N{BLACK SQUARE FOR STOP}", self.stop_pages),
            ("\N{INFORMATION SOURCE}", self.show_help),
        ]

        if ctx.guild is not None:
            self.permissions = self.channel.permissions_for(ctx.guild.me)
        else:
            self.permissions = self.channel.permissions_for(ctx.bot.user)

        if not self.permissions.embed_links:
            raise CannotPaginate("Bot does not have embed links permission.")

        if not self.permissions.send_messages:
            raise CannotPaginate("Bot cannot send messages.")

        if self.paginating:
            # verify we can actually use the pagination session
            if not self.permissions.add_reactions:
                raise CannotPaginate("Bot does not have add reactions permission.")

            if not self.permissions.read_message_history:
                raise CannotPaginate(
                    "Bot does not have Read Message History permission."
                )

    def get_page(self, page):
        base = (page - 1) * self.per_page
        return self.entries[base : base + self.per_page]

    def get_content(self, entries, page, *, first=False):
        return None

    def get_embed(self, entries, page, *, first=False):
        self.prepare_embed(entries, page, first=first)
        return self.embed

    def prepare_embed(self, entries, page, *, first=False):
        p = []
        for index, entry in enumerate(entries, 1 + ((page - 1) * self.per_page)):
            p.append(f"{index}. {entry}")

        if self.maximum_pages > 1:
            if self.show_entry_count:
                text = f"Page {page}/{self.maximum_pages} ({len(self.entries)} entries)"
            else:
                text = f"Page {page}/{self.maximum_pages}"

            self.embed.set_footer(text=text)

        if self.paginating and first:
            p.append("")
            p.append("Confused? React with \N{INFORMATION SOURCE} for more info.")

        self.embed.description = "\n".join(p)

    async def show_page(self, page, *, first=False):
        self.current_page = page
        entries = self.get_page(page)
        content = self.get_content(entries, page, first=first)
        embed = self.get_embed(entries, page, first=first)

        if not self.paginating:
            return await self.channel.send(content=content, embed=embed)

        if not first:
            await self.message.edit(content=content, embed=embed)
            return

        self.message = await self.channel.send(content=content, embed=embed)
        for (reaction, _) in self.reaction_emojis:
            if self.maximum_pages == 2 and reaction in ("\u23ed", "\u23ee"):
                # no |<< or >>| buttons if we only have two pages
                # we can't forbid it if someone ends up using it but remove
                # it from the default set
                continue

            await self.message.add_reaction(reaction)

    async def checked_show_page(self, page):
        if page != 0 and page <= self.maximum_pages:
            await self.show_page(page)

    async def first_page(self):
        """goes to the first page"""
        await self.show_page(1)

    async def last_page(self):
        """goes to the last page"""
        await self.show_page(self.maximum_pages)

    async def next_page(self):
        """goes to the next page"""
        await self.checked_show_page(self.current_page + 1)

    async def previous_page(self):
        """goes to the previous page"""
        await self.checked_show_page(self.current_page - 1)

    async def show_current_page(self):
        if self.paginating:
            await self.show_page(self.current_page)

    async def numbered_page(self):
        """lets you type a page number to go to"""
        to_delete = []
        to_delete.append(await self.channel.send("What page do you want to go to?"))

        def message_check(m):
            return (
                m.author == self.author
                and self.channel == m.channel
                and m.content.isdigit()
            )

        try:
            msg = await self.bot.wait_for("message", check=message_check, timeout=30.0)
        except asyncio.TimeoutError:
            to_delete.append(await self.channel.send("Took too long."))
            await asyncio.sleep(5)
        else:
            page = int(msg.content)
            to_delete.append(msg)
            if page != 0 and page <= self.maximum_pages:
                await self.show_page(page)
            else:
                to_delete.append(
                    await self.channel.send(
                        f"Invalid page given. ({page}/{self.maximum_pages})"
                    )
                )
                await asyncio.sleep(5)

        try:
            await self.channel.delete_messages(to_delete)
        except Exception:
            pass

    async def show_help(self):
        """shows this message"""
        messages = ["Welcome to the interactive paginator!\n"]
        messages.append(
            "This interactively allows you to see pages of text by navigating with "
            "reactions. They are as follows:\n"
        )

        for (emoji, func) in self.reaction_emojis:
            messages.append(f"{emoji} {func.__doc__}")

        embed = self.embed.copy()
        embed.clear_fields()
        embed.description = "\n".join(messages)
        embed.set_footer(
            text=f"We were on page {self.current_page} before this message."
        )
        await self.message.edit(content=None, embed=embed)

        async def go_back_to_current_page():
            await asyncio.sleep(60.0)
            await self.show_current_page()

        self.bot.loop.create_task(go_back_to_current_page())

    async def stop_pages(self):
        """stops the interactive pagination session"""
        await self.message.delete()
        self.paginating = False

    def react_check(self, payload):
        if payload.user_id != self.author.id:
            return False

        if payload.message_id != self.message.id:
            return False

        to_check = str(payload.emoji)
        for (emoji, func) in self.reaction_emojis:
            if to_check == emoji:
                self.match = func
                return True
        return False

    async def paginate(self):
        """Actually paginate the entries and run the interactive loop if necessary."""
        first_page = self.show_page(1, first=True)
        if not self.paginating:
            await first_page
        else:
            # allow us to react to reactions right away if we're paginating
            self.bot.loop.create_task(first_page)

        while self.paginating:
            try:
                payload = await self.bot.wait_for(
                    "raw_reaction_add", check=self.react_check, timeout=120.0
                )
            except asyncio.TimeoutError:
                self.paginating = False
                try:
                    await self.message.clear_reactions()
                except:
                    pass
                finally:
                    break

            try:
                await self.message.remove_reaction(
                    payload.emoji, discord.Object(id=payload.user_id)
                )
            except:
                pass  # can't remove it so don't bother doing so

            await self.match()



import asyncio
import datetime
from queue import PriorityQueue
from typing import Optional, Union

import discord
from discord import Member, PermissionOverwrite
from discord.ext import commands
from discord.ext.commands import BucketType

from plugins.debate.rooms.data import DebateRoom, Participant, Topic


def only_debate_channels():
    def predicate(ctx):
        return ctx.channel.id in ctx.cog.allowed_debate_channels

    return commands.check(predicate)


def only_command_channel():
    def predicate(ctx):
        return ctx.channel.id == ctx.cog.channels["tc_debate_feed"].id

    return commands.check(predicate)


def only_command_and_debate_channels():
    def predicate(ctx):
        channels = list(ctx.cog.allowed_debate_channels)
        channels.append(ctx.cog.channels["tc_debate_feed"].id)
        return ctx.channel.id in channels

    return commands.check(predicate)


def only_misc_channels():
    def predicate(ctx):
        return ctx.channel.id in ctx.cog.allowed_misc_channels

    return commands.check(predicate)


def only_misc_and_debate_channels():
    def predicate(ctx):
        return (
            ctx.channel.id
            in ctx.cog.allowed_misc_channels + ctx.cog.allowed_debate_channels
        )

    return commands.check(predicate)


def only_misc_debate_command_channels():
    def predicate(ctx):
        channels = list(ctx.cog.allowed_debate_channels)
        channels.append(ctx.cog.channels["tc_commands"].id)
        return ctx.channel.id in (ctx.cog.allowed_misc_channels + channels)

    return commands.check(predicate)


def disabled_while_concluding():
    def predicate(ctx):
        room_number = ctx.cog.get_room_number(ctx.channel)
        room = ctx.cog.get_room(room_number)

        if room is None:
            return True

        if room.match is None:
            return True

        if room.match.concluding:
            return False
        else:
            return True

    return commands.check(predicate)


class DebateRooms(commands.Cog, name="Debate"):
    def __init__(
        self,
        bot,
        guild,
        debate_room_maps: list,
        elo_role_maps: dict,
        allowed_debate_channels: list,
        allowed_misc_channels: list,
    ):
        self.bot = bot
        self.logger = self.bot.logger
        self.db = self.bot.db

        self.interface_messages = []
        self.exiting = False

        # Data
        self.guild = guild
        self.roles = self.bot.cogs["Server Setup"].roles
        self.debate_rooms = []
        self.debate_room_maps = debate_room_maps
        self.channels = self.bot.cogs["Server Setup"].channels
        self.elo_role_maps = elo_role_maps
        self.allowed_debate_channels = allowed_debate_channels
        self.allowed_misc_channels = allowed_misc_channels
        self.enabled = False

        # Implementation Junk
        self.debate_room_tcs = []

        # Room Visibility
        self.empty_rooms = PriorityQueue(maxsize=0)
        self.non_empty_rooms = PriorityQueue(maxsize=0)
        self.invisible_rooms = PriorityQueue(maxsize=0)
        self.visible_rooms = PriorityQueue(maxsize=0)
        self.empty_invisible_rooms = PriorityQueue(maxsize=0)
        self.empty_visible_rooms = PriorityQueue(maxsize=0)
        self.non_empty_invisible_rooms = PriorityQueue(maxsize=0)
        self.non_empty_visible_rooms = PriorityQueue(maxsize=0)

        # Tasks
        self.vc_visibility_task = self.bot.loop.create_task(self.voice_channel_update())

    @commands.Cog.listener()
    async def on_ready(self):
        await self.debates_enabled()

    async def debates_enabled(self):
        for mapping in self.debate_room_maps:
            tc = self.bot.get_channel(mapping[0])
            vc = self.bot.get_channel(mapping[1])
            self.debate_rooms.append(
                DebateRoom(
                    self.debate_room_maps.index(mapping) + 1,
                    tc,  # Text Channel
                    vc,  # Voice Channel
                )
            )

        self.debate_room_tcs = [room.tc for room in self.debate_rooms]
        await self.delete_recent_messages()
        for room in self.debate_rooms:
            message = await self.send_embed_message(room.number)
            self.interface_messages.append(message)

        self.enabled = True

    # Convenience Methods
    def get_room_number(
        self, channel: Union[discord.TextChannel, discord.VoiceChannel]
    ):
        """Get a room number from a TextChannel or VoiceChannel ID."""
        try:
            number = list(
                filter(lambda x: x.number_from_channel(channel), self.debate_rooms)
            )[0].number
            return number
        except IndexError as e_info:
            return None

    def get_room(self, room_num: int):
        """Get a room from a room number."""
        try:
            room = list(filter(lambda x: x.number == room_num, self.debate_rooms))[0]
            return room
        except IndexError as e_info:
            return None

    def get_vc_from_tc(self, tc: discord.TextChannel) -> discord.VoiceChannel:
        """Get a VoiceChannel from a TextChannel."""
        vc = list(filter(lambda x: x.tc == tc, self.debate_rooms))[0].vc
        return vc

    def get_tc_from_vc(self, vc: discord.VoiceChannel) -> discord.TextChannel:
        """Get a TextChannel from a Voice Channel."""
        tc = list(filter(lambda x: x.vc == vc, self.debate_rooms))[0].tc
        return tc

    def check_debate_vc(self, vc: discord.VoiceChannel) -> bool:
        """Ensure VoiceChannel is a debate room."""
        return vc in [room.vc for room in self.debate_rooms]

    # End of Convenience Methods

    @commands.Cog.listener()
    async def on_message(self, message):
        if message.author == self.bot.user:
            # Do nothing if the message is persistent embed
            if len(message.embeds) > 0:
                if message.embeds[0].title.startswith("Debate Room"):
                    return

        if message.channel in self.debate_room_tcs:
            # Get number
            room_num = self.get_room_number(message.channel)

            # Delete interface message
            index = room_num - 1
            im_del = self.interface_messages[index]
            try:
                await im_del.delete()
            except discord.errors.NotFound as e_info:
                return

    @commands.Cog.listener()
    async def on_message_delete(self, message):
        if message.channel in self.debate_room_tcs:
            # Add interface message when embed is deleted
            if message.id in [i.id for i in self.interface_messages]:
                index = self.get_room_number(message.channel) - 1
                if not self.exiting:
                    im = await self.add_interface_message(index)

    @commands.Cog.listener()
    async def on_bulk_message_delete(self, messages):
        for message in messages:
            if message.channel in self.debate_room_tcs:
                # Add interface message when embed is deleted
                if message.id in [i.id for i in self.interface_messages]:
                    index = self.get_room_number(message.channel) - 1
                    if not self.exiting:
                        im = await self.add_interface_message(index)

    async def delete_recent_messages(self):
        """Delete most recent messages sent by the bot account."""
        for room in self.debate_rooms:
            tc = room.tc
            async for message in tc.history(limit=100):
                if message.author == self.bot.user:
                    await message.delete()

    def get_embed_message(self, room_num):
        response = discord.Embed(
            color=0xEB6A5C,
            title=f"Debate Room {room_num}",
            description="Set channel topics democratically and vote for "
            "the best debater. Start a debate to unmute yourself.",
        )
        return response

    async def send_embed_message(self, room_num):
        embed = self.get_embed_message(room_num)

        topic_updated = self.get_room(room_num).set_current_topic()
        current_topic = self.get_room(room_num).current_topic
        if current_topic:
            embed.add_field(name="**Topic**: ", value=f"{current_topic}")
        text_channel = self.bot.get_channel(list(self.debate_room_tcs)[room_num - 1].id)
        message = await text_channel.send(embed=embed)
        return message

    async def add_interface_message(self, index):
        room_num = index + 1
        im_add = await self.send_embed_message(room_num)
        self.interface_messages[index] = im_add
        return self.interface_messages

    async def update_im(self, room_num):
        index = room_num - 1
        im = self.interface_messages[index]
        embed = self.get_embed_message(room_num)
        topic = self.get_room(room_num).current_topic
        if topic:
            embed.add_field(
                name="**Topic**: ", value=f"{self.get_room(room_num).current_topic}"
            )
        try:
            await im.edit(embed=embed)
        except discord.errors.NotFound as e_info:
            return

    def check_in_debate(self, ctx):
        """Check if someone is in the corresponding TextChannel or
        VoiceChannel to allow a command.
        """
        tc = ctx.message.channel
        if not (tc in [room.tc for room in self.debate_rooms]):
            return False
        vc = self.get_vc_from_tc(tc)
        if ctx.message.author.voice:
            author_vc = ctx.message.author.voice.channel
        else:
            return False
        return vc == author_vc

    async def conclude_debate(self, room, debaters):
        for debater in debaters:
            # Mute
            if debater.member in room.vc.members:
                await debater.member.edit(mute=True)

        for debater in debaters:
            await room.vc.set_permissions(debater.member, overwrite=None)

        embed = discord.Embed(
            title="⏸ Debate concluding..",
            description="ELO ratings are being updated. "
            "Debate specific commands will not run.",
        )

        check_voters = room.match.check_voters()
        if not check_voters:
            debaters = []

        if debaters:
            await room.tc.send(embed=embed, delete_after=60)

        for debater in debaters:
            await self.db.upsert(debater.member, elo=debater.elo_post)

            debate_feed = self.channels["tc_debate_feed"]
            embed = discord.Embed(
                title="ELO Rating Change",
            )
            embed.set_footer(
                text=debater.member.display_name,
                icon_url=debater.member.avatar_url,
            )
            embed.add_field(name="Before: ", value=f"{debater.elo_pre}")
            embed.add_field(name="After: ", value=f"{debater.elo_post}")
            await debate_feed.send(embed=embed)

            for key, val in self.elo_role_maps.items():
                guild = room.vc.guild
                if guild.get_member(debater):
                    await debater.member.remove_roles(
                        debater.member.guild.get_role(self.elo_role_maps[key]),
                        reason="Automatically removed at end of match.",
                    )

            for key, val in self.elo_role_maps.items():
                if key < debater.elo_post:
                    guild = room.vc.guild
                    if guild.get_member(debater):
                        await debater.member.add_roles(
                            debater.member.guild.get_role(self.elo_role_maps[key]),
                            reason="Automatically added at the end of a match.",
                        )
                    break

        embed = discord.Embed(
            title="✅ Debate concluded.",
            description="ELO ratings have been updated.",
        )
        if not check_voters:
            embed.description = (
                "ELO ratings have not been updated due to lack of voters."
            )
        room.match = None  # Clear match
        if debaters:
            await room.tc.send(embed=embed)

    async def update_topic(self, room):
        """Update topic of room from current topic."""
        if len(room.vc.members) == 0:
            current_topic = None
            topic_updated = False
        else:
            topic_updated = room.set_current_topic()
            current_topic = room.current_topic

        match = room.match

        self.logger.debug(f"Topics: {room.topics}")
        if current_topic is not None:
            if not match:
                topic_updated = room.set_current_topic()
                current_topic = room.current_topic
                room.start_match(current_topic)
                await self.update_im(room.number)
            else:
                for member in room.vc.members:
                    await room.vc.set_permissions(member, overwrite=None)

                if not room.match:
                    return

                # Do nothing if there are no voters
                if not room.match.check_voters():
                    debaters = room.match.get_debaters()

                    if not topic_updated:
                        return

                    # Mute debaters early
                    for debater in debaters:
                        # Remove overwrite from VC and mute
                        await room.vc.set_permissions(debater.member, overwrite=None)
                        if debater.member in room.vc.members:
                            await debater.member.edit(mute=True)
                    return

                debaters = []
                if match.concluding is False and match.concluded is False:
                    if topic_updated:

                        debaters = room.stop_match()

                        # Mute debaters early
                        for debater in debaters:
                            # Remove overwrite from VC and mute
                            await room.vc.set_permissions(
                                debater.member, overwrite=None
                            )
                            if debater.member in room.vc.members:
                                await debater.member.edit(mute=True)

                        match.concluding = True
                        await self.conclude_debate(room, debaters)
                        match.concluding = False
                        match.concluded = True
                elif match.concluding is False and match.concluded is True:
                    return
                elif match.concluding is True and match.concluded is False:
                    return

                topic_updated = room.set_current_topic()
                current_topic = room.current_topic
                room.start_match(current_topic)
                await self.update_im(room.number)
                for member in room.vc.members:
                    await member.edit(mute=True)

        topic_updated = room.set_current_topic()
        current_topic = room.current_topic
        await self.update_im(room.number)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="topic",
        brief="Set or vote for a topic in a debate room.",
        help="Set a topic in an empty debate room if you're the first setter. "
        "If you're not the first to set the topic, then vote on any "
        "user's topic or propose your own. A successful topic change "
        "will cause the ELO ratings to be calculated for debaters.",
    )
    async def topic(
        self, ctx, member: Union[Member, str], *, message: commands.clean_content = ""
    ):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if isinstance(member, Member):

            if self.check_debate_vc(self.get_vc_from_tc(ctx.channel)):
                topic = room.vote_topic(voter=ctx.author, candidate=member)
        else:
            if len(str(message)) > 1024:
                embed = discord.Embed(title="❌ Topic is longer than 1024 characters! ❌")
                await ctx.channel.send(embed=embed, delete_after=20)
                return
            else:
                topic_updated = room.add_topic(
                    Topic(
                        member=ctx.author,
                        message=f"{member} {str(message)}",  # Look carefully
                    )
                )
                if topic_updated:
                    embed = discord.Embed(
                        title="⚠️ Votes on your topic have been reset "
                        "because you updated it! ⚠️"
                    )
                    await ctx.channel.send(embed=embed, delete_after=10)
        await self.update_topic(room)

    async def make_vc_visible(self, vc):
        await vc.edit(sync_permissions=True)

    async def make_vc_invisible(self, vc):
        overwrite = PermissionOverwrite(view_channel=False)
        await vc.set_permissions(self.roles["role_citizen"], overwrite=overwrite)
        await vc.set_permissions(self.roles["role_member"], overwrite=overwrite)

    def vc_is_visible(self, vc):
        return vc.permissions_synced

    def vc_is_empty(self, vc):
        return len(vc.members) == 0

    async def voice_channel_update(self):
        await self.bot.wait_until_ready()
        while not self.bot.is_closed():
            self.empty_rooms = PriorityQueue(maxsize=0)
            self.non_empty_rooms = PriorityQueue(maxsize=0)
            self.invisible_rooms = PriorityQueue(maxsize=0)
            self.visible_rooms = PriorityQueue(maxsize=0)
            self.empty_invisible_rooms = PriorityQueue(maxsize=0)
            self.empty_visible_rooms = PriorityQueue(maxsize=0)
            self.non_empty_invisible_rooms = PriorityQueue(maxsize=0)
            self.non_empty_visible_rooms = PriorityQueue(maxsize=0)

            for room in self.debate_rooms:
                if self.vc_is_empty(room.vc):
                    self.empty_rooms.put(room)
                    if not self.vc_is_visible(room.vc):
                        self.empty_invisible_rooms.put(room)
                    else:
                        self.empty_visible_rooms.put(room)
                else:
                    self.non_empty_rooms.put(room)
                    if not self.vc_is_visible(room.vc):
                        self.non_empty_invisible_rooms.put(room)
                    else:
                        self.non_empty_visible_rooms.put(room)

                if not self.vc_is_visible(room.vc):
                    self.invisible_rooms.put(room)
                else:
                    self.visible_rooms.put(room)

            if self.empty_visible_rooms.empty():
                empty_invisible_room = self.empty_invisible_rooms.get()
                await self.make_vc_visible(empty_invisible_room.vc)
                self.visible_rooms.put(empty_invisible_room)
            else:
                if self.empty_visible_rooms.qsize() > 1:
                    for room in self.empty_visible_rooms.queue[1:]:
                        await self.make_vc_invisible(room.vc)

            await asyncio.sleep(1)

    @commands.Cog.listener()
    async def on_voice_state_update(self, member, before, after):

        # Variables
        dr_vcs = [dr.vc for dr in self.debate_rooms]

        # Do nothing if mute event
        if before.mute and not after.mute:
            return
        if not before.mute and after.mute:
            return

        if before.deaf and not after.deaf:
            return
        if not before.deaf and after.deaf:
            return

        if before.self_mute and not after.self_mute:
            return
        if not before.self_mute and after.self_mute:
            return

        if before.self_deaf and not after.self_deaf:
            return
        if not before.self_deaf and after.self_deaf:
            return

        async def join_room():
            self.logger.debug(f"{member} joined: {after.channel}")
            room_after = self.get_room(self.get_room_number(after.channel))
            room_after.add_topic_voter(member)
            room_after.reset_topic_creation(member)

            if room_after.match:
                participant = room_after.match.get_participant(member)
                if participant:
                    participant.session_start = datetime.datetime.utcnow()

            # Make linked text chat visible
            tc_after = self.get_tc_from_vc(after.channel)
            overwrite = PermissionOverwrite(read_messages=True)
            await tc_after.set_permissions(member, overwrite=overwrite)

            await member.edit(mute=True)

        async def leave_room():
            self.logger.debug(f"{member} left: {before.channel}")
            room_before = self.get_room(self.get_room_number(before.channel))
            room_before.remove_topic_voter(member)
            room_before.remove_priority_from_topic(member)
            room_before.remove_obsolete_topic(member)

            active_debaters = []
            if room_before.match:
                participant = room_before.match.get_participant(member)
                if participant:
                    participant.session_end = datetime.datetime.utcnow()
                    participant.update_duration()

                debaters = [d.member for d in room_before.match.get_debaters()]
                for voice_member in before.channel.members:
                    if voice_member in debaters:
                        active_debaters.append(voice_member)

                self.logger.debug(
                    f"Active Debaters (On Leave Room):"
                    f" {[a.name for a in active_debaters]}"
                )
                self.logger.debug(
                    f"# of Active Debaters (On Leave Room): {len(active_debaters)}"
                )

                if len(active_debaters) <= 1:
                    if room_before.match.concluding or room_before.match.concluded:
                        pass
                    else:
                        room_before.match.concluding = True
                        await self.conclude_debate(
                            room_before, debaters=room_before.stop_match()
                        )

            # Remove overwrite from VC
            await room_before.vc.set_permissions(member, overwrite=None)

            # Make linked text chat invisible
            tc_before = self.get_tc_from_vc(before.channel)
            await tc_before.set_permissions(member, overwrite=None)

            # Delete if not working
            await self.update_topic(room_before)

        async def switch_room():
            # Join Room
            self.logger.debug(f"{member} joined: {after.channel}")
            room_after = self.get_room(self.get_room_number(after.channel))
            room_after.add_topic_voter(member)
            room_after.reset_topic_creation(member)

            # Make linked text chat visible
            tc_after = self.get_tc_from_vc(after.channel)
            overwrite = PermissionOverwrite(read_messages=True)
            await tc_after.set_permissions(member, overwrite=overwrite)

            # Leave Room
            self.logger.debug(f"{member} left: {before.channel}")
            room_before = self.get_room(self.get_room_number(before.channel))
            room_before.remove_topic_voter(member)
            room_before.remove_priority_from_topic(member)
            room_before.remove_obsolete_topic(member)

            active_debaters = []
            if room_before.match:
                participant = room_before.match.get_participant(member)
                if participant:
                    participant.session_end = datetime.datetime.utcnow()
                    participant.update_duration()

                debaters = [d.member for d in room_before.match.get_debaters()]
                for voice_member in before.channel.members:
                    if voice_member in debaters:
                        active_debaters.append(voice_member)

                self.logger.debug(
                    f"Active Debaters (On Switch Out): "
                    f"{[a.name for a in active_debaters]}"
                )
                self.logger.debug(
                    f"# of Active Debaters (On Switch Out): {len(active_debaters)}"
                )

                if len(active_debaters) <= 1:
                    if room_before.match.concluding or room_before.match.concluded:
                        pass
                    else:
                        room_before.match.concluding = True
                        await self.conclude_debate(
                            room_before, debaters=room_before.stop_match()
                        )

            # Remove overwrite from VC
            await room_before.vc.set_permissions(member, overwrite=None)

            # Make linked text chat invisible
            tc_before = self.get_tc_from_vc(before.channel)
            await tc_before.set_permissions(member, overwrite=None)

            # Mute User
            await member.edit(mute=True)

            # Delete if not working
            await self.update_topic(room_before)

        # If member joins a debate room
        if before.channel is None and after.channel in dr_vcs:
            await join_room()
            return
        # If member leaves a debate room
        if before.channel in dr_vcs and after.channel is None:
            await leave_room()
            return

        if after.channel:
            after_list = list(dr_vcs)
            after_list.remove(after.channel)
        else:
            return

        if before.channel:
            before_list = list(dr_vcs)
            before_list.remove(before.channel)
        else:
            return

        if before.channel in after_list and after.channel in before_list:
            await switch_room()
            return

    @commands.Cog.listener()
    async def on_member_join(self, member):
        if member.bot:
            return
        await self.db.upsert(member, elo=1500)
        await member.add_roles(
            member.guild.get_role(self.elo_role_maps[800]),
            reason="Automatically added since a new user.",
        )

    @commands.Cog.listener()
    async def on_member_remove(self, member):
        document = await self.db[self.db.database].MemberStates.find_one(
            {"member_id": member.id}
        )
        if document:
            await self.db[self.db.database].MemberStates.delete_one(
                {"member_id": member.id}
            )

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.has_any_role("Staff", "Director", "Moderator")
    @commands.command(
        name="remove-topic",
        brief="Remove a topic from a debate room.",
        help="This command will remove the topic a user has proposed. If an "
        "active topic is removed then ELO ratings for that debate will be "
        "calculated.",
    )
    async def remove_topic(self, ctx, member: Optional[Member]):
        room_num = self.get_room_number(ctx.channel)
        room = None
        if room_num:
            room = self.get_room(room_num)
        if room:
            if member:
                topic = room.topic_from_member(member)
            else:
                topic = room.current_topic
            if room.match:
                if topic == room.match.topic:
                    for current_member in room.vc.members:
                        await current_member.edit(mute=True)
            if member:
                room.remove_topic(member)
            else:
                room.remove_topic(room.current_topic.author)
            room.match = None  # Clear match
            await self.update_topic(room)

    @commands.has_role("Engineering")
    @commands.command(
        name="hard-reset-elo-ratings",
        brief="Permanently reset the ELO ratings of users.",
        help="This command will purge the database of all ELO ratings and "
        "give them a default rating of 1500.",
    )
    async def reset_scores(self, ctx):
        guild = ctx.guild

        # Count humans
        count = 0
        for member in guild.members:
            if member.bot:
                continue

            count += 1

        response = discord.Embed(color=0x696969, title=f"🔍 Processing {count} members.")
        message = await ctx.send(embed=response)

        count = 0
        for member in guild.members:
            if member.bot:
                continue

            await self.db.upsert(member, elo=1500)
            count += 1

            for key, val in self.elo_role_maps.items():
                await member.remove_roles(
                    member.guild.get_role(self.elo_role_maps[key]),
                    reason="Automatically removed during reset.",
                )

            title = f"🔍 Processed {count} members."
            if count % 5 == 0:
                response = discord.Embed(color=0x696969, title=title)
                await message.edit(embed=response)

        response = discord.Embed(
            color=0x77B255, title="✅ All ELO scores have been reset."
        )
        await message.edit(embed=response, delete_after=10)

    @commands.has_role("Engineering")
    @commands.command(
        name="setup-elo",
        brief="Initialize ELO ratings in a new server.",
        help="This command will give the default ELO rating to users and "
        "setup default initialization procedures for ELO ratings.",
    )
    async def setup_elo(self, ctx):
        guild = ctx.guild

        # Count humans
        count = 0
        for member in guild.members:
            if member.bot:
                continue

            count += 1

        response = discord.Embed(color=0x696969, title=f"🔍 Processing {count} members.")
        message = await ctx.send(embed=response)
        count = 0
        for member in guild.members:
            if member.bot:
                continue
            await self.db.upsert(member, elo=1500)
            count += 1

            for key, val in self.elo_role_maps.items():
                await member.remove_roles(
                    member.guild.get_role(self.elo_role_maps[key]),
                    reason="Automatically removed during setup.",
                )

            await member.add_roles(
                member.guild.get_role(self.elo_role_maps[800]),
                reason="Automatically added during setup.",
            )

            title = f"🔍 Processed {count} members."
            if count % 5 == 0:
                response = discord.Embed(color=0x696969, title=title)
                await message.edit(embed=response)

        response = discord.Embed(
            color=0x77B255,
            title="✅ ELO ratings have been setup and roles have been assigned.",
        )
        await message.edit(embed=response, delete_after=10)

    @commands.has_role("Engineering")
    @commands.command(
        name="repair-elo",
        brief="Repairs the ELO ratings of buggy users.",
        help="This command will check all users to see if they have missing "
        "ELO ratings, fix their roles and update the database.",
    )
    async def repair_elo(self, ctx):
        guild = ctx.guild

        # Count humans
        count = 0
        for member in guild.members:
            if member.bot:
                continue

            count += 1

        response = discord.Embed(color=0x696969, title=f"🔍 Processing {count} members.")
        message = await ctx.send(embed=response)
        count = 0
        for member in guild.members:
            if member.bot:
                continue
            count += 1
            elo = await self.db.get(member, state="elo")
            if elo:
                for key, val in self.elo_role_maps.items():
                    await member.remove_roles(
                        member.guild.get_role(self.elo_role_maps[key]),
                        reason="Automatically removed during setup.",
                    )

                for key, val in self.elo_role_maps.items():
                    if key < elo:
                        await member.add_roles(
                            member.guild.get_role(self.elo_role_maps[key]),
                            reason="Automatically added during setup.",
                        )
                        break
            else:
                await self.db.upsert(member, elo=1500)
                for key, val in self.elo_role_maps.items():
                    await member.remove_roles(
                        member.guild.get_role(self.elo_role_maps[key]),
                        reason="Automatically removed during setup.",
                    )

                await member.add_roles(
                    member.guild.get_role(self.elo_role_maps[800]),
                    reason="Automatically added during setup.",
                )

            title = f"🔍 Processed {count} members."
            if count % 5 == 0:
                response = discord.Embed(color=0x696969, title=title)
                await message.edit(embed=response)

        response = discord.Embed(
            color=0x77B255,
            title="✅ Missing ELO ratings and roles " "have been assigned.",
        )
        await message.edit(embed=response, delete_after=30)

    @only_misc_debate_command_channels()
    @commands.cooldown(1, 5, BucketType.user)
    @commands.command(
        name="elo",
        brief="Display ELO rating of a user.",
        help="This command will display the current ELO score of a user.",
    )
    async def elo(self, ctx, member: Optional[discord.Member]):
        if member:
            if member.bot:
                return
        if member:
            elo = await self.db.get(member, state="elo")
        else:
            elo = await self.db.get(ctx.author, state="elo")
        if not elo:
            return
        embed = discord.Embed(
            title="Debate ELO Rating",
            description=f"`{str(elo)}`",
        )
        if member:
            embed.set_footer(text=member.display_name, icon_url=member.avatar_url)
        else:
            embed.set_footer(
                text=ctx.author.display_name, icon_url=ctx.author.avatar_url
            )
        await ctx.send(embed=embed)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="for",
        brief="Set the 'For' position on a topic in a debate room.",
        help="This command will allow you to vote on a debate. You are 'For' "
        "the topic.",
    )
    async def command_for(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate running right now. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        elo = await self.db.get(ctx.author, state="elo")

        room.match.add_for(
            Participant(
                member=ctx.author, elo=elo, session_start=datetime.datetime.utcnow()
            )
        )

        embed = discord.Embed(title="✅ You are now for the position.")
        await ctx.send(embed=embed, delete_after=10)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="against",
        brief="Set the 'Against' position on a topic in a debate room.",
        help="This command will allow you to vote on a debate. You are 'Against' "
        "the topic.",
    )
    async def against(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        elo = await self.db.get(ctx.author, state="elo")

        room.match.add_against(
            Participant(
                member=ctx.author, elo=elo, session_start=datetime.datetime.utcnow()
            )
        )

        embed = discord.Embed(title="✅ You are now against the position.")
        await ctx.send(embed=embed, delete_after=10)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="debate",
        brief="Start or join a debate.",
        help="This command will allow you to start or join an existing "
        "debate. You must have already selected a position on a topic "
        "for this command to have any effect.",
    )
    async def debate(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        if not room.match.check_participant(ctx.author):
            embed = discord.Embed(
                title="You must choose a position on the topic before "
                "you can debate.",
                description="`$for` - For the topic.\n\n"
                "`$against` - Against the topic.",
            )
            await ctx.send(embed=embed, delete_after=60)
            return

        current_session_start = datetime.datetime.utcnow()
        for participant in room.match.participants:
            participant.session_start = current_session_start

        room.match.add_debater(ctx.author)

        embed = discord.Embed(
            title="✅ You are now a debater on the topic.",
            description="Your ELO rating is at risk. " "Be mindful of what you say.",
        )
        await ctx.send(embed=embed, delete_after=20)
        if self.roles["role_muted"] in ctx.author.roles:
            await ctx.author.edit(mute=True)
        else:
            await ctx.author.edit(mute=False)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="debate-for",
        aliases=["df"],
        brief="Shortcut to debate for the topic.",
        help="This command has the same affect as taking the 'For' position on a "
        "topic and then starting or joining a debate.",
    )
    async def debate_for(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate running right now. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        elo = await self.db.get(ctx.author, state="elo")

        room.match.add_for(
            Participant(
                member=ctx.author, elo=elo, session_start=datetime.datetime.utcnow()
            )
        )

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        current_session_start = datetime.datetime.utcnow()
        for participant in room.match.participants:
            participant.session_start = current_session_start

        room.match.add_debater(ctx.author)

        embed = discord.Embed(
            title="✅ You are now a debater for the topic.",
            description="Your ELO rating is at risk. Be mindful of what you say.",
        )
        await ctx.send(embed=embed, delete_after=20)
        if self.roles["role_muted"] in ctx.author.roles:
            await ctx.author.edit(mute=True)
        else:
            await ctx.author.edit(mute=False)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="debate-against",
        aliases=["da"],
        brief="Shortcut to debate against the topic.",
        help="This command has the same affect as taking the 'Against' position "
        "on a "
        "topic and then starting or joining a debate.",
    )
    async def debate_against(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        elo = await self.db.get(ctx.author, state="elo")

        room.match.add_against(
            Participant(
                member=ctx.author, elo=elo, session_start=datetime.datetime.utcnow()
            )
        )

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        current_session_start = datetime.datetime.utcnow()
        for participant in room.match.participants:
            participant.session_start = current_session_start

        room.match.add_debater(ctx.author)

        embed = discord.Embed(
            title="✅ You are now a debater against the topic.",
            description="Your ELO rating is at risk. " "Be mindful of what you say.",
        )
        await ctx.send(embed=embed, delete_after=20)
        if self.roles["role_muted"] in ctx.author.roles:
            await ctx.author.edit(mute=True)
        else:
            await ctx.author.edit(mute=False)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="vote",
        brief="Vote for who you think won the debate.",
        help="This command will cast a vote for a debater in an active "
        "debate. You can switch votes before the end of a debate.",
    )
    async def vote(self, ctx, member: discord.Member):
        room = self.get_room(self.get_room_number(ctx.channel))

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        if member == ctx.author:
            embed = discord.Embed(title="❌ You cannot vote for yourself dummy. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        result = room.match.vote(voter=ctx.author, candidate=member)
        if result is None:
            embed = discord.Embed(
                title="❌ You can only vote for debaters if you take a position. ❌"
            )
            await ctx.send(embed=embed, delete_after=10)
            return
        if result is False:
            embed = discord.Embed(title="❌ You can only vote for debaters. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return
        embed = discord.Embed(title="✅ Your vote has been cast.")
        await ctx.send(embed=embed, delete_after=20)

    @only_debate_channels()
    @disabled_while_concluding()
    @commands.command(
        name="conclude",
        brief="Conclude an active debate.",
        help="This command will end an active debate in a debate room. It "
        "will also cause the calculation of ELO ratings and remove the "
        "topic from the room.",
    )
    async def conclude(self, ctx):
        room = self.get_room(self.get_room_number(ctx.channel))
        match = room.match

        # Exit if not in a debate room
        if not self.check_in_debate(ctx):
            return

        if not room.check_match():
            embed = discord.Embed(title="❌ No debate match in progress. ❌")
            await ctx.send(embed=embed, delete_after=10)
            return

        debaters, concluded, voters = room.vote_conclude(voter=ctx.author)
        self.logger.debug(f"Debaters (On Conclude): {debaters}")
        self.logger.debug(f"Concluded (On Conclude): {concluded}")

        # Do nothing if another conclude command already run
        if concluded is None:
            return
        else:
            embed = discord.Embed(title="✅ Vote to conclude debate cast.")
            await ctx.send(embed=embed, delete_after=10)

        if concluded:
            match.concluding = True
            embed = discord.Embed(
                title="⏸ Debate concluding..",
                description="ELO ratings are being updated.",
            )
            await ctx.send(embed=embed, delete_after=60)
            for debater in debaters:
                # Mute
                if debater.member in room.vc.members:
                    await debater.member.edit(mute=True)

            for debater in debaters:
                await room.vc.set_permissions(debater.member, overwrite=None)

            if voters:
                for debater in debaters:
                    # Update database
                    await self.db.upsert(debater.member, elo=debater.elo_post)

                    debate_feed = self.channels["tc_debate_feed"]
                    embed = discord.Embed(
                        title="ELO Rating Change",
                    )
                    embed.set_footer(
                        text=debater.member.display_name,
                        icon_url=debater.member.avatar_url,
                    )
                    embed.add_field(name="Before: ", value=f"{debater.elo_pre}")
                    embed.add_field(name="After: ", value=f"{debater.elo_post}")
                    await debate_feed.send(embed=embed)

                    # Remove ELO roles
                    for key, val in self.elo_role_maps.items():
                        guild = room.vc.guild
                        if guild.get_member(debater):
                            await debater.member.remove_roles(
                                debater.member.guild.get_role(self.elo_role_maps[key]),
                                reason="Automatically removed at end of match.",
                            )

                    # Add ELO roles
                    for key, val in self.elo_role_maps.items():
                        if key < debater.elo_post:
                            guild = room.vc.guild
                            if guild.get_member(debater):
                                await debater.member.add_roles(
                                    debater.member.guild.get_role(
                                        self.elo_role_maps[key]
                                    ),
                                    reason="Automatically added at the "
                                    "end of a match.",
                                )
                            break

            # Update topic
            current_topic = room.current_topic
            if current_topic:
                await self.update_im(self.get_room_number(room.tc))
                room.remove_topic(current_topic.author)
                room.vote_topic(current_topic.author, current_topic.author)
            else:
                await self.update_im(self.get_room_number(room.tc))

            # Remove voters from data set
            room.remove_conclude_voters()
            room.match = None  # Clear match

            await self.update_topic(room)

            # Send conclude message
            embed = discord.Embed(
                title="✅ Debate concluded.",
                description="ELO ratings have been updated.",
            )
            await ctx.send(embed=embed)
            match.concluding = False
            match.concluded = True

        else:
            if debaters:
                if len(debaters) == 0:
                    embed = discord.Embed(
                        title="❌ Debate not concluded. ❌",
                        description="You cannot conclude a debate with 0 "
                        "participants.",
                    )
                    await ctx.send(embed=embed)

    async def lockdown_cancel_all_matches(self):
        for room in self.debate_rooms:
            if room.match:
                room.match = None
            room.purge_topics()
            index = room.number - 1
            await self.interface_messages[index].edit(
                embed=self.get_embed_message(room_num=room.number)
            )

            for member in room.vc.members:
                await member.move_to(None)
                await room.vc.set_permissions(member, overwrite=None)

                # Make linked text chat invisible
                await room.tc.set_permissions(member, overwrite=None)

            overwrite = PermissionOverwrite(view_channel=False)
            if room.number >= 2:
                await room.vc.set_permissions(
                    self.roles["role_member"], overwrite=overwrite
                )
                await room.vc.set_permissions(
                    self.roles["role_citizen"], overwrite=overwrite
                )

    async def debates_disabled(self):
        self.exiting = True
        for message in self.interface_messages:
            try:
                await message.delete()
            except discord.errors.NotFound as e_info:
                self.bot.logger.debug(
                    f"Interface message during exit process could not be deleted."
                )
        self.enabled = False



import asyncio

import discord
from discord import ChannelType, Colour, PermissionOverwrite, Permissions, utils
from discord.ext import commands
from discord.ext.commands import BucketType

from plugins.debate.overwrites import (
    _debate_tc_citizen_permissions,
    _debate_tc_member_permissions,
    generate_overwrite,
    lockdown_citizen_general_perms,
    lockdown_debate_member_perms,
    lockdown_permissions,
)
from plugins.debate.rooms.interface import DebateRooms


class ServerSetup(commands.Cog, name="Server Setup"):
    def __init__(self, bot):
        self.bot = bot
        self.logger = self.bot.logger
        self.db = self.bot.db

        # Roles
        self.role_warden = None
        self.role_engineering = None
        self.role_staff = None
        self.role_director = None
        self.role_moderator = None
        self.role_grandmaster = None
        self.role_legend = None
        self.role_master = None
        self.role_expert = None
        self.role_distinguished = None
        self.role_apprentice = None
        self.role_novice = None
        self.role_initiate = None
        self.role_rookie = None
        self.role_incompetent = None
        self.role_bot = None
        self.role_citizen = None
        self.role_member = None
        self.role_events = None
        self.role_logs = None
        self.role_detained = None
        self.role_muted = None
        self.role_super_muted = None
        self.role_everyone = None

        self.roles = {
            "role_warden": self.role_warden,
            "role_engineering": self.role_engineering,
            "role_staff": self.role_staff,
            "role_director": self.role_director,
            "role_moderator": self.role_moderator,
            "role_grandmaster": self.role_grandmaster,
            "role_legend": self.role_legend,
            "role_master": self.role_master,
            "role_expert": self.role_expert,
            "role_distinguished": self.role_distinguished,
            "role_apprentice": self.role_apprentice,
            "role_novice": self.role_novice,
            "role_initiate": self.role_initiate,
            "role_rookie": self.role_rookie,
            "role_incompetent": self.role_incompetent,
            "role_bot": self.role_bot,
            "role_citizen": self.role_citizen,
            "role_member": self.role_member,
            "role_events": self.role_events,
            "role_logs": self.role_logs,
            "role_detained": self.role_detained,
            "role_muted": self.role_muted,
            "role_super_muted": self.role_super_muted,
        }

        # Channels
        self.category_information = None
        self.channel_rules = None
        self.channel_about = None
        self.channel_verification = None
        self.channel_announcements = None
        self.channel_community_updates = None
        self.category_moderation = None
        self.channel_director_commands = None
        self.channel_mod_commands = None
        self.channel_isolation = None
        self.category_interface = None
        self.channel_debate_feed = None
        self.channel_commands = None
        self.category_community = None
        self.channel_general = None
        self.channel_serious = None
        self.channel_meme_dump = None
        self.category_debate = None
        self.category_logs = None
        self.channel_moderator_actions = None
        self.channel_message_deletion = None
        self.channel_message_edits = None
        self.channel_ban_unban = None
        self.channel_nicknames = None
        self.channel_join_leave = None
        self.channel_automod = None
        self.channel_channels = None
        self.channel_invites = None
        self.channel_roles = None
        self.channel_voice = None

        self.channels = {
            "category_information": self.category_information,
            "tc_rules": self.channel_rules,
            "tc_about": self.channel_about,
            "tc_verification": self.channel_verification,
            "tc_announcements": self.channel_announcements,
            "tc_community_updates": self.channel_community_updates,
            "category_moderation": self.category_moderation,
            "tc_director_commands": self.channel_director_commands,
            "tc_mod_commands": self.channel_mod_commands,
            "tc_isolation": self.channel_isolation,
            "category_interface": self.category_interface,
            "tc_debate_feed": self.channel_debate_feed,
            "tc_commands": self.channel_commands,
            "category_community": self.category_community,
            "tc_general": self.channel_general,
            "tc_serious": self.channel_serious,
            "tc_meme_dump": self.channel_meme_dump,
            "category_debate": self.category_debate,
            "category_logs": self.category_logs,
            "tc_moderator_actions": self.channel_moderator_actions,
            "tc_message_deletion": self.channel_message_deletion,
            "tc_message_edits": self.channel_message_edits,
            "tc_ban_unban": self.channel_ban_unban,
            "tc_nicknames": self.channel_nicknames,
            "tc_join_leave": self.channel_join_leave,
            "tc_automod": self.channel_automod,
            "tc_channels": self.channel_channels,
            "tc_invites": self.channel_invites,
            "tc_roles": self.channel_roles,
            "tc_voice": self.channel_voice,
        }

        # Debate
        self.debate_rooms = []
        self.debate_room_maps = []
        self.elo_role_maps = {}

    async def cog_check(self, ctx):
        """Check if user has the Engineering role."""
        engineering = utils.get(ctx.guild.roles, name="Engineering")
        return engineering in ctx.author.roles

    @commands.cooldown(1, 50000, BucketType.guild)
    @commands.command(
        name="setup-roles",
        brief="Set up roles required by the bot in the server.",
        help="This command will set up all roles and the associated "
        "permissions required by the bot in the server. This command "
        "should not be used multiple times in a period of 24 hrs as you "
        "will be rate limited.",
    )
    async def setup_roles(self, ctx):
        """Setup roles for the server."""
        guild = ctx.guild

        # Indicate command is running.
        title = f"🔍 Processing Roles"
        response = discord.Embed(color=0x696969, title=title)
        progress_message = await ctx.channel.send(embed=response)

        # Delete All Roles
        for role in guild.roles[1:]:
            if not role.managed:
                await role.delete(reason="Server Setup")
                await asyncio.sleep(5)

        # Setup Power Roles
        self.role_everyone = guild.default_role
        await self.role_everyone.edit(permissions=Permissions(permissions=0))

        self.role_warden = await guild.create_role(
            name="Warden",
            permissions=Permissions(permissions=8),
            color=Colour(value=0xEB6A5C),
            hoist=False,
        )
        await ctx.me.add_roles(self.role_warden)

        self.role_engineering = await guild.create_role(
            name="Engineering", permissions=Permissions(permissions=8), hoist=False
        )
        await guild.owner.add_roles(self.role_engineering)

        self.role_staff = await guild.create_role(
            name="Staff", permissions=Permissions(permissions=36768832), hoist=False
        )
        await guild.owner.add_roles(self.role_staff)

        self.role_director = await guild.create_role(
            name="Director",
            permissions=Permissions(permissions=1275064288),
            color=Colour(value=0xE74C3C),
            hoist=False,
        )
        await guild.owner.add_roles(self.role_director)

        self.role_moderator = await guild.create_role(
            name="Moderator",
            permissions=Permissions(permissions=267775936),
            color=Colour(value=0x2ECC71),
            hoist=False,
        )

        # Setup Rated Roles
        self.role_grandmaster = await guild.create_role(
            name="Grandmaster 👑",
            permissions=Permissions(permissions=0),
            hoist=True,
        )

        self.role_legend = await guild.create_role(
            name="Legend 🏆", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_master = await guild.create_role(
            name="Master ⚖️", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_expert = await guild.create_role(
            name="Expert ⚔️", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_distinguished = await guild.create_role(
            name="Distinguished 💥",
            permissions=Permissions(permissions=0),
            hoist=True,
        )

        self.role_apprentice = await guild.create_role(
            name="Apprentice 💡",
            permissions=Permissions(permissions=0),
            hoist=True,
        )

        self.role_novice = await guild.create_role(
            name="Novice 🔥", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_initiate = await guild.create_role(
            name="Initiate 🔰", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_rookie = await guild.create_role(
            name="Rookie 🧷", permissions=Permissions(permissions=0), hoist=True
        )

        self.role_incompetent = await guild.create_role(
            name="Incompetent 💯",
            permissions=Permissions(permissions=0),
            hoist=True,
        )

        self.role_bot = await guild.create_role(
            name="Bot", permissions=Permissions(permissions=0), hoist=True
        )
        await ctx.me.add_roles(self.role_bot)

        # Setup Basic Roles
        self.role_citizen = await guild.create_role(
            name="Citizen", permissions=Permissions(permissions=104189504), hoist=False
        )

        self.role_member = await guild.create_role(
            name="Member", permissions=Permissions(permissions=36768832), hoist=False
        )

        # Setup Miscellaneous Roles
        self.role_events = await guild.create_role(
            name="Events", permissions=Permissions(permissions=0), hoist=False
        )

        self.role_logs = await guild.create_role(
            name="Logs", permissions=Permissions(permissions=0), hoist=False
        )

        # Setup Punishment Roles
        self.role_detained = await guild.create_role(
            name="Detained", permissions=Permissions(permissions=0), hoist=False
        )

        self.role_muted = await guild.create_role(
            name="Muted", permissions=Permissions(permissions=0), hoist=False
        )

        self.role_super_muted = await guild.create_role(
            name="Super Muted", permissions=Permissions(permissions=0), hoist=False
        )

        # Update Database
        await self.db.upsert(
            guild,
            role_warden=self.role_warden.id,
            role_engineering=self.role_engineering.id,
            role_staff=self.role_staff.id,
            role_director=self.role_director.id,
            role_moderator=self.role_moderator.id,
            role_grandmaster=self.role_grandmaster.id,
            role_legend=self.role_legend.id,
            role_master=self.role_master.id,
            role_expert=self.role_expert.id,
            role_distinguished=self.role_distinguished.id,
            role_apprentice=self.role_apprentice.id,
            role_novice=self.role_novice.id,
            role_initiate=self.role_initiate.id,
            role_rookie=self.role_rookie.id,
            role_incompetent=self.role_incompetent.id,
            role_bot=self.role_bot.id,
            role_citizen=self.role_citizen.id,
            role_member=self.role_member.id,
            role_events=self.role_events.id,
            role_logs=self.role_logs.id,
            role_detained=self.role_detained.id,
            role_muted=self.role_muted.id,
            role_super_muted=self.role_super_muted.id,
        )

        # Confirm database has been updated.
        response = discord.Embed(color=0x77B255, title="✅ Roles Updated")
        await progress_message.edit(embed=response, delete_after=30)

    async def check_roles(self, ctx, message):
        guild = ctx.guild
        title = f"❌ {message} ❌"
        for role in self.roles.keys():
            # Check if role exists within database.
            role_id = await self.db.get(guild, state=f"{role}")
            if role_id:
                # Check if role exists within server.
                grabbed_role = guild.get_role(role_id)
                self.roles[f"{role}"] = grabbed_role
                if grabbed_role is None:
                    embed = discord.Embed(title=title)
                    await ctx.send(embed=embed, delete_after=10)
                    return False
            else:
                embed = discord.Embed(title=title)
                await ctx.send(embed=embed, delete_after=10)
                return False
        return True

    async def check_channels(self, ctx, message):
        guild = ctx.guild
        title = f"❌ {message} ❌"
        for channel in self.channels.keys():
            # Check if channel exists within database.
            channel_id = await self.db.get(guild, state=f"{channel}")
            if channel_id:
                # Check if channel exists within server.
                grabbed_channel = guild.get_channel(channel_id)
                self.channels[f"{channel}"] = grabbed_channel
                if grabbed_channel is None:
                    embed = discord.Embed(title=title)
                    await ctx.send(embed=embed, delete_after=10)
                    return False
            else:
                embed = discord.Embed(title=title)
                await ctx.send(embed=embed, delete_after=10)
                return False

        for _channel_number in range(1, 21):
            tc_id = await self.db.get(guild, state=f"tc_debate_{_channel_number}")

            if tc_id:
                grabbed_channel = guild.get_channel(tc_id)
                if grabbed_channel is None:
                    embed = discord.Embed(title=title)
                    await ctx.send(embed=embed, delete_after=10)
                    return False
            else:
                embed = discord.Embed(title=title)
                await ctx.send(embed=embed, delete_after=10)
                return False

            vc_id = await self.db.get(guild, state=f"vc_debate_{_channel_number}")
            if vc_id:
                grabbed_channel = guild.get_channel(vc_id)
                if grabbed_channel is None:
                    embed = discord.Embed(title=title)
                    await ctx.send(embed=embed, delete_after=10)
                    return False
            else:
                embed = discord.Embed(title=title)
                await ctx.send(embed=embed, delete_after=10)
                return False
        return True

    @commands.command(
        name="setup-channels",
        brief="Set up channels required by the bot in the server.",
        help="This command will set up all channels and it's permissions "
        "required by the bot in the server.",
    )
    async def setup_channels(self, ctx):
        guild = ctx.guild
        self.role_everyone = guild.default_role

        # Indicate command is running.
        title = f"🔍 Processing Channels"
        response = discord.Embed(color=0x696969, title=title)
        progress_message = await ctx.channel.send(embed=response)

        roles_set_up = await self.check_roles(
            ctx, message="Cannot set up channels without roles set up."
        )

        # Exit early if roles not set up
        if not roles_set_up:
            return

        # Delete All Channels
        await guild.rules_channel.edit(category=None)
        await guild.public_updates_channel.edit(category=None)

        for channel in guild.channels:
            skipped_channels = [guild.rules_channel, guild.public_updates_channel]

            if channel not in skipped_channels:
                await channel.delete(reason="Server Setup")

        # Setup Information Category
        self.category_information = await guild.create_category(
            name="Information",
            overwrites=generate_overwrite(ctx, self.roles, "information"),
        )

        self.channel_rules = guild.rules_channel

        await guild.rules_channel.edit(
            name="rules", category=self.category_information, sync_permissions=True
        )

        self.channel_about = await guild.create_text_channel(
            name="about", category=self.category_information, sync_permissions=True
        )

        self.channel_verification = await guild.create_text_channel(
            name="verification",
            category=self.category_information,
            overwrites=generate_overwrite(ctx, self.roles, "verification"),
        )

        self.channel_announcements = await guild.create_text_channel(
            name="announcements",
            category=self.category_information,
            sync_permissions=True,
        )
        await self.channel_announcements.edit(type=ChannelType.news)

        self.channel_community_updates = guild.public_updates_channel
        await guild.public_updates_channel.edit(
            name="community-updates",
            category=self.category_information,
            position=8,
            overwrites=generate_overwrite(ctx, self.roles, "community_updates"),
        )

        # Setup Moderation Category
        self.category_moderation = await guild.create_category(
            name="Moderation",
            overwrites=generate_overwrite(ctx, self.roles, "moderation"),
        )

        self.channel_director_commands = await guild.create_text_channel(
            name="director-commands",
            category=self.category_moderation,
            overwrites=generate_overwrite(ctx, self.roles, "director_commands"),
        )

        self.channel_mod_commands = await guild.create_text_channel(
            name="mod-commands",
            category=self.category_moderation,
            sync_permissions=True,
        )

        self.channel_isolation = await guild.create_text_channel(
            name="isolation",
            category=self.category_moderation,
            overwrites=generate_overwrite(ctx, self.roles, "isolation"),
            slowmode_delay=15,
        )

        # Setup Interface Category
        self.category_interface = await guild.create_category(
            name="Interface",
            overwrites=generate_overwrite(ctx, self.roles, "interface"),
        )

        self.channel_debate_feed = await guild.create_text_channel(
            name="debate-feed", category=self.category_interface, sync_permissions=True
        )

        self.channel_commands = await guild.create_text_channel(
            name="commands",
            category=self.category_interface,
            overwrites=generate_overwrite(ctx, self.roles, "commands"),
            slowmode_delay=5,
        )

        # Setup Community Category
        self.category_community = await guild.create_category(
            name="Community",
            overwrites=generate_overwrite(ctx, self.roles, "community"),
        )

        self.channel_general = await guild.create_text_channel(
            name="general", category=self.category_community, sync_permissions=True
        )

        self.channel_serious = await guild.create_text_channel(
            name="serious", category=self.category_community, sync_permissions=True
        )

        self.channel_meme_dump = await guild.create_text_channel(
            name="meme-dump", category=self.category_community, sync_permissions=True
        )

        # Setup Debate Category
        self.category_debate = await guild.create_category(
            name="Debate", overwrites=generate_overwrite(ctx, self.roles, "debate")
        )

        # Setup Logs Category
        self.category_logs = await guild.create_category(
            name="Logs", overwrites=generate_overwrite(ctx, self.roles, "logs")
        )

        self.channel_moderator_actions = await guild.create_text_channel(
            name="moderator-actions", category=self.category_logs, sync_permissions=True
        )

        self.channel_message_deletion = await guild.create_text_channel(
            name="message-deletion", category=self.category_logs, sync_permissions=True
        )

        self.channel_message_edits = await guild.create_text_channel(
            name="message-edits", category=self.category_logs, sync_permissions=True
        )

        self.channel_ban_unban = await guild.create_text_channel(
            name="ban-unban", category=self.category_logs, sync_permissions=True
        )

        self.channel_nicknames = await guild.create_text_channel(
            name="nicknames", category=self.category_logs, sync_permissions=True
        )

        self.channel_join_leave = await guild.create_text_channel(
            name="join-leave", category=self.category_logs, sync_permissions=True
        )

        self.channel_automod = await guild.create_text_channel(
            name="automod", category=self.category_logs, sync_permissions=True
        )

        self.channel_channels = await guild.create_text_channel(
            name="channels", category=self.category_logs, sync_permissions=True
        )

        self.channel_invites = await guild.create_text_channel(
            name="invites", category=self.category_logs, sync_permissions=True
        )

        self.channel_roles = await guild.create_text_channel(
            name="roles", category=self.category_logs, sync_permissions=True
        )

        self.channel_voice = await guild.create_text_channel(
            name="voice", category=self.category_logs, sync_permissions=True
        )

        # Create Debate Channels
        for _channel_number in range(1, 21):
            _tc_debate = await guild.create_text_channel(
                name=f"debate-{_channel_number}",
                category=self.category_debate,
                overwrites=generate_overwrite(ctx, self.roles, "debate-#"),
            )

            _vc_debate = await guild.create_voice_channel(
                name=f"Debate {_channel_number}", category=self.category_debate
            )

            overwrite = PermissionOverwrite(view_channel=False)
            if _channel_number != 1:
                await _vc_debate.set_permissions(
                    self.roles["role_citizen"], overwrite=overwrite
                )

                await _vc_debate.set_permissions(
                    self.roles["role_member"], overwrite=overwrite
                )

            _channel_ids = {
                f"tc_debate_{_channel_number}": _tc_debate.id,
                f"vc_debate_{_channel_number}": _vc_debate.id,
            }
            await self.db.upsert(guild, **_channel_ids)

        # Update Database
        await self.db.upsert(
            guild,
            category_information=self.category_information.id,
            tc_rules=guild.rules_channel.id,
            tc_about=self.channel_about.id,
            tc_verification=self.channel_verification.id,
            tc_announcements=self.channel_announcements.id,
            tc_community_updates=guild.public_updates_channel.id,
            category_moderation=self.category_moderation.id,
            tc_director_commands=self.channel_director_commands.id,
            tc_mod_commands=self.channel_mod_commands.id,
            tc_isolation=self.channel_isolation.id,
            category_interface=self.category_interface.id,
            tc_debate_feed=self.channel_debate_feed.id,
            tc_commands=self.channel_commands.id,
            category_community=self.category_community.id,
            tc_general=self.channel_general.id,
            tc_serious=self.channel_serious.id,
            tc_meme_dump=self.channel_meme_dump.id,
            category_debate=self.category_debate.id,
            category_logs=self.category_logs.id,
            tc_moderator_actions=self.channel_moderator_actions.id,
            tc_message_deletion=self.channel_message_deletion.id,
            tc_message_edits=self.channel_message_edits.id,
            tc_ban_unban=self.channel_ban_unban.id,
            tc_nicknames=self.channel_nicknames.id,
            tc_join_leave=self.channel_join_leave.id,
            tc_automod=self.channel_automod.id,
            tc_channels=self.channel_channels.id,
            tc_invites=self.channel_invites.id,
            tc_roles=self.channel_roles.id,
            tc_voice=self.channel_voice.id,
        )

        # Confirm database has been updated.
        response = discord.Embed(color=0x77B255, title="✅ Channels Updated")
        await progress_message.edit(embed=response, delete_after=30)

    @commands.command(
        name="enable-debates",
        brief="Enable the 'Debates' cog.",
        help="This command will check if all the required channels and roles "
        "are available for the bot to use. If it is successful in "
        "confirming the same, then it will enable the 'Debates' cog.",
    )
    async def enable_debates(self, ctx):
        # Indicate command is running.
        title = f"🔍 Enabling Debates"
        response = discord.Embed(color=0x696969, title=title)
        progress_message = await ctx.channel.send(embed=response)

        if "Debate" in self.bot.cogs:
            if self.bot.cogs["Debate"].enabled:
                response = discord.Embed(title="❌ Debates Already Enabled ❌")
                await ctx.send(embed=response)
                return

        roles_set_up = await self.check_roles(
            ctx, message="Cannot enable debates without roles set up."
        )

        # Exit early if roles are not set up
        if not roles_set_up:
            await progress_message.delete()
            return

        channels_set_up = await self.check_channels(
            ctx, message="Cannot enable debates without channels set up."
        )

        if not channels_set_up:
            await progress_message.delete()
            return

        guild = ctx.guild

        for _channel_number in range(1, 21):
            _tc_debate = await self.db.get(guild, state=f"tc_debate_{_channel_number}")

            _vc_debate = await self.db.get(guild, state=f"vc_debate_{_channel_number}")

            self.debate_room_maps.append((_tc_debate, _vc_debate))

        elo_ratings = [2800, 2600, 2400, 2200, 2000, 1800, 1600, 800, 400, 100]
        role_ids = []
        for role in list(self.roles.keys())[5:15]:
            role_ids.append(self.roles[role].id)
        self.elo_role_maps = dict(zip(elo_ratings, role_ids))

        allowed_debate_channels = [i for (i, k) in self.debate_room_maps]
        allowed_misc_channels = [
            self.channel_director_commands,
            self.channel_mod_commands,
            self.channel_commands,
        ]

        self.bot.add_cog(
            DebateRooms(
                self.bot,
                guild,
                self.debate_room_maps,
                self.elo_role_maps,
                allowed_debate_channels,
                allowed_misc_channels,
            )
        )

        await self.bot.cogs["Debate"].debates_enabled()

        # Confirm Debates is Enabled
        response = discord.Embed(color=0x77B255, title="✅ Debates Enabled")
        await progress_message.edit(embed=response, delete_after=30)

    @commands.command(
        name="disable-debates",
        brief="Disable the 'Debates' cog.",
        help="This command will disable the 'Debates' cog and launch all exit "
        "mechanisms associated with the same cog.",
    )
    async def disable_debates(self, ctx):
        # Indicate command is running.
        title = f"🔍 Disabling Debates"
        response = discord.Embed(color=0x696969, title=title)
        progress_message = await ctx.channel.send(embed=response)

        if "Debate" in self.bot.cogs:
            if self.bot.cogs["Debate"].enabled:
                await self.bot.cogs["Debate"].debates_disabled()

            self.bot.remove_cog("Debate")

        # Confirm debates is enabled
        response = discord.Embed(color=0x77B255, title="✅ Debates Disabled")
        await progress_message.edit(embed=response, delete_after=30)

    @commands.has_role("Staff")
    @commands.command(
        name="lockdown",
        brief="Locks the entire server down during downtime issues.",
        help="This commands locks down the whole server from being used"
        "by all members except staff. Should only be used when all of the "
        "moderation bots go down.",
    )
    async def lockdown(self, ctx):
        await self.channels["tc_commands"].set_permissions(
            self.roles["role_member"], overwrite=lockdown_permissions
        )
        await self.channels["tc_commands"].set_permissions(
            self.roles["role_citizen"], overwrite=lockdown_permissions
        )
        await self.channels["category_community"].set_permissions(
            self.roles["role_member"], overwrite=lockdown_permissions
        )
        await self.channels["category_community"].set_permissions(
            self.roles["role_citizen"], overwrite=lockdown_permissions
        )

        await self.channels["category_debate"].set_permissions(
            self.roles["role_member"], overwrite=lockdown_debate_member_perms
        )

        await self.channels["category_debate"].set_permissions(
            self.roles["role_citizen"], overwrite=lockdown_citizen_general_perms
        )

        for _channel_num in range(1, 21):
            vc = discord.utils.get(
                ctx.guild.voice_channels, name=f"Debate {_channel_num}"
            )
            tc = discord.utils.get(
                ctx.guild.text_channels, name=f"debate-{_channel_num}"
            )
            await tc.set_permissions(
                self.roles["role_member"],
                create_instant_invite=False,
                manage_channels=False,
                add_reactions=False,
                read_messages=False,
                view_channel=False,
                send_messages=False,
                send_tts_messages=False,
                manage_messages=False,
                embed_links=False,
                attach_files=False,
                read_message_history=True,
                mention_everyone=False,
                external_emojis=False,
                manage_permissions=False,
                manage_webhooks=False,
            )
            await tc.set_permissions(
                self.roles["role_citizen"],
                create_instant_invite=False,
                manage_channels=False,
                add_reactions=False,
                read_messages=False,
                view_channel=False,
                send_messages=False,
                send_tts_messages=False,
                manage_messages=False,
                embed_links=False,
                attach_files=False,
                read_message_history=True,
                mention_everyone=False,
                external_emojis=False,
                manage_permissions=False,
                manage_webhooks=False,
            )

            if _channel_num != 1:
                await vc.set_permissions(
                    self.roles["role_member"],
                    priority_speaker=False,
                    stream=False,
                    view_channel=False,
                    connect=False,
                    speak=False,
                    mute_members=False,
                    deafen_members=False,
                    move_members=False,
                    use_voice_activation=False,
                    manage_permissions=False,
                )
                await vc.set_permissions(
                    self.roles["role_citizen"],
                    priority_speaker=False,
                    stream=False,
                    view_channel=False,
                    connect=False,
                    speak=False,
                    mute_members=False,
                    deafen_members=False,
                    move_members=False,
                    use_voice_activation=False,
                    manage_permissions=False,
                )
            else:
                await vc.set_permissions(
                    self.roles["role_member"],
                    priority_speaker=False,
                    stream=False,
                    view_channel=True,
                    connect=False,
                    speak=False,
                    mute_members=False,
                    deafen_members=False,
                    move_members=False,
                    use_voice_activation=False,
                    manage_permissions=False,
                )
                await vc.set_permissions(
                    self.roles["role_citizen"],
                    priority_speaker=False,
                    stream=False,
                    view_channel=True,
                    connect=False,
                    speak=False,
                    mute_members=False,
                    deafen_members=False,
                    move_members=False,
                    use_voice_activation=False,
                    manage_permissions=False,
                )

        await self.bot.cogs["Debate"].lockdown_cancel_all_matches()

    @commands.has_role("Director")
    @commands.command(
        name="reopen",
        brief="Unlocks the server for public use.",
        help="This commands unlocks the server for public use. It should only be used "
        "when at least one of the moderation bots are back up. Only directors can"
        "use this command.",
    )
    async def reopen(self, ctx):
        await self.channels["tc_commands"].edit(
            overwrites=generate_overwrite(ctx, self.roles, "commands")
        )
        await self.channels["category_community"].edit(
            overwrites=generate_overwrite(ctx, self.roles, "community")
        )
        await self.channels["tc_general"].edit(sync_permissions=True)
        await self.channels["tc_serious"].edit(sync_permissions=True)
        await self.channels["tc_meme_dump"].edit(sync_permissions=True)
        await self.channels["category_debate"].edit(
            overwrites=generate_overwrite(ctx, self.roles, "debate")
        )

        for _channel_num in range(1, 21):
            vc = discord.utils.get(
                ctx.guild.voice_channels, name=f"Debate {_channel_num}"
            )
            tc = discord.utils.get(
                ctx.guild.text_channels, name=f"debate-{_channel_num}"
            )
            await vc.edit(sync_permissions=True)
            overwrite = PermissionOverwrite(view_channel=False)
            if _channel_num != 1:
                await vc.set_permissions(
                    self.roles["role_citizen"], overwrite=overwrite
                )

                await vc.set_permissions(self.roles["role_member"], overwrite=overwrite)
            await tc.set_permissions(
                self.roles["role_member"], overwrite=_debate_tc_member_permissions
            )
            await tc.set_permissions(
                self.roles["role_citizen"], overwrite=_debate_tc_citizen_permissions
            )

    @commands.has_role("Engineering")
    @commands.command(
        name="quit",
        brief="Quit the bot and logout of bot account.",
        help="This command will disable all associated cogs, quit the "
        "process and log out of the bot account.",
    )
    async def quit(self, ctx):
        if "Debate" in self.bot.cogs:
            pass
        else:
            await self.bot.logout()
        if self.bot.cogs["Debate"].enabled:
            await self.bot.cogs["Debate"].debates_disabled()
        await self.bot.logout()



from discord import PermissionOverwrite

_staff_team_permissive = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=True,
    add_reactions=True,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=True,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=True,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=True,
    deafen_members=True,
    move_members=True,
    use_voice_activation=True,
    manage_permissions=True,
    manage_webhooks=False,
)

_no_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=False,
    view_channel=False,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=False,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

lockdown_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_neutral_read_only_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=None,
    view_channel=None,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_read_only_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)


_muted_read_only_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_moderation_send_all_and_read_all_permission = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=True,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_moderation_privileged_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=True,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=True,
    deafen_members=True,
    move_members=True,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_staff_team_read_only = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=True,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_read_and_send_only = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_citizen_general_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

lockdown_citizen_general_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=True,
    connect=False,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_moderator_general_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=True,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=True,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=True,
    deafen_members=True,
    move_members=True,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)


_senior_mod_general_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=True,
    stream=True,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=True,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=True,
    external_emojis=True,
    connect=True,
    speak=True,
    mute_members=True,
    deafen_members=True,
    move_members=True,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_moderator_log_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_senior_mod_log_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=False,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=True,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_debate_member_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=True,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

lockdown_debate_member_perms = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=False,
    read_messages=True,
    view_channel=True,
    send_messages=False,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=True,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=True,
    manage_permissions=False,
    manage_webhooks=False,
)

_debate_tc_citizen_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=False,
    read_messages=False,
    view_channel=False,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=True,
    attach_files=True,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=True,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_debate_tc_member_permissions = PermissionOverwrite(
    create_instant_invite=False,
    manage_channels=False,
    add_reactions=True,
    priority_speaker=False,
    stream=False,
    read_messages=False,
    view_channel=False,
    send_messages=True,
    send_tts_messages=False,
    manage_messages=False,
    embed_links=False,
    attach_files=False,
    read_message_history=True,
    mention_everyone=False,
    external_emojis=False,
    connect=False,
    speak=False,
    mute_members=False,
    deafen_members=False,
    move_members=False,
    use_voice_activation=False,
    manage_permissions=False,
    manage_webhooks=False,
)

_neutral_permissions = PermissionOverwrite(
    create_instant_invite=None,
    manage_channels=None,
    add_reactions=None,
    priority_speaker=None,
    stream=None,
    read_messages=None,
    view_channel=None,
    send_messages=None,
    send_tts_messages=None,
    manage_messages=None,
    embed_links=None,
    attach_files=None,
    read_message_history=None,
    mention_everyone=None,
    external_emojis=None,
    connect=None,
    speak=None,
    mute_members=None,
    deafen_members=None,
    move_members=None,
    use_voice_activation=None,
    manage_permissions=None,
    manage_webhooks=None,
)

all_channel_overwrites = {
    "information": {
        "role_warden": _staff_team_permissive,
        "role_director": PermissionOverwrite(
            create_instant_invite=False,
            manage_channels=False,
            add_reactions=False,
            priority_speaker=False,
            stream=False,
            read_messages=True,
            view_channel=True,
            send_messages=False,
            send_tts_messages=False,
            manage_messages=True,
            embed_links=True,
            attach_files=True,
            read_message_history=True,
            mention_everyone=True,
            external_emojis=True,
            connect=False,
            speak=False,
            mute_members=False,
            deafen_members=False,
            move_members=False,
            use_voice_activation=False,
            manage_permissions=False,
            manage_webhooks=False,
        ),
        "role_moderator": _neutral_read_only_permissions,
        "role_citizen": _neutral_read_only_permissions,
        "role_member": _neutral_read_only_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _neutral_read_only_permissions,
        "role_muted": _neutral_read_only_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _neutral_permissions,
    },
    "verification": {
        "role_warden": _staff_team_permissive,
        "role_director": _staff_team_read_only,
        "role_moderator": _no_permissions,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": PermissionOverwrite(
            create_instant_invite=False,
            manage_channels=False,
            add_reactions=False,
            priority_speaker=False,
            stream=False,
            read_messages=True,
            view_channel=True,
            send_messages=False,
            send_tts_messages=False,
            manage_messages=False,
            embed_links=False,
            attach_files=False,
            read_message_history=True,
            mention_everyone=False,
            external_emojis=False,
            connect=False,
            speak=False,
            mute_members=False,
            deafen_members=False,
            move_members=False,
            use_voice_activation=True,
            manage_permissions=False,
            manage_webhooks=False,
        ),
    },
    "community_updates": {
        "role_warden": _staff_team_permissive,
        "role_director": _staff_team_read_only,
        "role_moderator": _no_permissions,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
    "moderation": {
        "role_warden": _staff_team_permissive,
        "role_director": _moderation_privileged_permissions,
        "role_moderator": _moderation_send_all_and_read_all_permission,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
    "director_commands": {
        "role_warden": _staff_team_permissive,
        "role_director": _moderation_privileged_permissions,
        "role_moderator": _no_permissions,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
    "isolation": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _moderator_general_perms,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _read_and_send_only,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
    "interface": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _neutral_read_only_permissions,
        "role_citizen": _neutral_read_only_permissions,
        "role_member": _neutral_read_only_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _neutral_read_only_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _neutral_permissions,
    },
    "commands": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _moderator_general_perms,
        "role_citizen": _neutral_permissions,
        "role_member": _neutral_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _neutral_read_only_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _neutral_permissions,
    },
    "community": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _moderator_general_perms,
        "role_citizen": _neutral_permissions,
        "role_member": _neutral_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _neutral_read_only_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _neutral_permissions,
    },
    "debate": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _moderator_general_perms,
        "role_citizen": _neutral_permissions,
        "role_member": _neutral_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _muted_read_only_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _neutral_permissions,
    },
    "debate-#": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_general_perms,
        "role_moderator": _moderator_general_perms,
        "role_citizen": _debate_tc_citizen_permissions,
        "role_member": _debate_tc_member_permissions,
        "role_logs": _neutral_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
    "logs": {
        "role_warden": _staff_team_permissive,
        "role_director": _senior_mod_log_perms,
        "role_moderator": _moderator_log_permissions,
        "role_citizen": _no_permissions,
        "role_member": _no_permissions,
        "role_logs": _read_only_permissions,
        "role_detained": _no_permissions,
        "role_muted": _no_permissions,
        "role_super_muted": _no_permissions,
        "role_everyone": _no_permissions,
    },
}


def generate_overwrite(ctx, roles, channel: str):
    """
    Parameters
    ----------
    ctx: discord context
    roles: a dict of roles pulled from db
    channel: name of any GuildChannel

    Returns
    -------
    A dict of roles and their overwrites for a specific channel
    """
    overwrites = all_channel_overwrites[channel]
    return {
        roles["role_warden"]: overwrites["role_warden"],
        roles["role_director"]: overwrites["role_director"],
        roles["role_moderator"]: overwrites["role_moderator"],
        roles["role_citizen"]: overwrites["role_citizen"],
        roles["role_member"]: overwrites["role_member"],
        roles["role_logs"]: overwrites["role_logs"],
        roles["role_detained"]: overwrites["role_detained"],
        roles["role_muted"]: overwrites["role_muted"],
        roles["role_super_muted"]: overwrites["role_super_muted"],
        ctx.guild.default_role: overwrites["role_everyone"],
    }
